#include "window.h"
#include <QApplication>
#include <QStandardPaths>
#include <QDebug>
#include <QThread>
#include <QProgressBar>
#include <QSlider>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QSpacerItem>
#include <QTimer>

Window::Window(QWidget *parent)
    : QWidget{parent}
{
    m_solveButton = new QPushButton("solve", this);
    connect(m_solveButton, SIGNAL(clicked()), this, SLOT(solveButtonClicked()));
    m_clearButton = new QPushButton("clear", this);
    connect(m_clearButton, SIGNAL(clicked()), this, SLOT(clearButtonClicked()));

    m_slider = new QSlider(this);
    m_slider->setRange(5, 2000);
    m_slider->setOrientation(Qt::Horizontal);
    m_slider->setValue(m_delay);
    connect(m_slider, SIGNAL(valueChanged(int)), this, SLOT(setValue(int)));

    QHBoxLayout *horizontalLayout = new QHBoxLayout(this);
    QVBoxLayout *verticalLayout = new QVBoxLayout();
    QGridLayout *layout = new QGridLayout();

    m_spacer[0] = new QSpacerItem(4, 4, QSizePolicy::Minimum, QSizePolicy::Fixed);
    m_spacer[1] = new QSpacerItem(4, 4, QSizePolicy::Minimum, QSizePolicy::Fixed);
    m_spacer[2] = new QSpacerItem(8, 8, QSizePolicy::Minimum, QSizePolicy::Expanding);
    verticalLayout->addItem(m_spacer[2]);

    verticalLayout->addWidget(m_solveButton);
    verticalLayout->addWidget(m_slider);
    verticalLayout->addWidget(m_clearButton);
    horizontalLayout->addLayout(layout, 5);
    horizontalLayout->addLayout(verticalLayout, 1);

    layout->addItem(m_spacer[0], 3, 3);
    layout->addItem(m_spacer[1], 7, 7);
    layout->setSpacing(0);

    int dx = 0;
    int dy = 0;
    for (int x = 0; x < 9; x++)
    {
        for (int y = 0; y < 9; y++)
        {
            m_cell[x + y * 9].setParent(this);
            layout->addWidget(&m_cell[x + y * 9], y + y / 3, x + x / 3);
            // connecting signals & slots
            for (int d = 0; d < 9; d++) // rows and columns
            {
                connect(&m_cell[x + y * 9], SIGNAL(update(int)), &m_cell[x + d * 9], SLOT(removeOption(int)));
                connect(&m_cell[x + y * 9], SIGNAL(update(int)), &m_cell[d + y * 9], SLOT(removeOption(int)));

                connect(&m_cell[x + y * 9], SIGNAL(undo(int)), &m_cell[x + d * 9], SLOT(addOption(int)));
                connect(&m_cell[x + y * 9], SIGNAL(undo(int)), &m_cell[d + y * 9], SLOT(addOption(int)));
            }
            int a = x / 3 * 3;
            int b = y / 3 * 3;
            for (int x1 = a; x1 < a + 3; x1++) // square
            {
                for (int y1 = b; y1 < b + 3; y1++)
                {
                    if (x1 == x || y1 == y)
                        continue;
                    connect(&m_cell[x + y * 9], SIGNAL(update(int)), &m_cell[x1 + y1 * 9], SLOT(removeOption(int)));

                    connect(&m_cell[x + y * 9], SIGNAL(undo(int)), &m_cell[x1 + y1 * 9], SLOT(addOption(int)));
                }
            }
        }
    }
}

void Window::solveButtonClicked()
{
    std::list<History> hist;
    bool backtracking = false;
    srand(time(NULL)); // random seed
    while (1)
    {
        repaint(); // draw cells again
        delay(m_delay);
        if (backtracking)
        {
            hist.back().cell->highlight("#0000ff");
            repaint(); // draw cells again
            delay(m_delay);
            hist.back().cell->clearHighlight();
            hist.back().cell->un();
            if (hist.back().index.size() == 0)
            {
                hist.pop_back();
                continue;
            }
            backtracking = false;
            int choiceNr = rand() % hist.back().index.size(); // choose a random number;
            repaint();                                         // draw cells again
            delay(m_delay);
            hist.back().cell->collapse(hist.back().index[choiceNr]); // collapse cell with chosen number;
            std::vector<int>::iterator it;
            it = hist.back().index.begin() + choiceNr;
            hist.back().index.erase(it);
        }
        else
        {
            std::vector<Cell *> b; // vector of cells with least entropy
            int minEtropy = 10;
            for (int x = 0; x < 9; x++) // loop through all cells and fill list b --> this could be optimzed
            {
                for (int y = 0; y < 9; y++)
                {
                    if (m_cell[x + y * 9].collapsed)
                        continue;
                    else if (m_cell[x + y * 9].possibleStates < minEtropy)
                    {
                        minEtropy = m_cell[x + y * 9].possibleStates;
                        b.clear();
                        b.push_back(&m_cell[x + y * 9]);
                    }
                    else if (m_cell[x + y * 9].possibleStates == minEtropy)
                    {
                        b.push_back(&m_cell[x + y * 9]);
                    }
                }
            }
            if (b.size() == 0) // if b is empty -> solved
            {
                qInfo("Solved!!");
                return;
            }
            else if (minEtropy == 0)
            {
                qInfo("backtracking");
                for (auto &c : b)
                {
                    c->highlight("#ff0000");
                }
                repaint(); // draw cells again
                delay(m_delay);
                for (auto &c : b)
                {
                    c->clearHighlight();
                }
                backtracking = true;
                continue;
            }
            for (auto &c : b)
            {
                c->highlight("#aaaa00");
            }
            repaint(); // draw cells again
            delay(m_delay);
            for (auto &c : b)
            {
                c->clearHighlight();
            }
            int choiceCell = rand() % b.size();                  // choose a random cell
            int choiceNr = rand() % b[choiceCell]->index.size(); // choose a random number;
            hist.push_back(History(b[choiceCell], b[choiceCell]->index, choiceNr));
            b[choiceCell]->highlight("#00ff00");
            repaint(); // draw cells again
            delay(m_delay);
            b[choiceCell]->clearHighlight();
            b[choiceCell]->collapse(b[choiceCell]->index[choiceNr]); // collapse that random cell with chosen number;
        }
    }
    // qInfo("%d---:)");
}
inline void Window::delay(int millisecondsWait)
{
    QEventLoop loop;
    QTimer t;
    t.connect(&t, &QTimer::timeout, &loop, &QEventLoop::quit);
    t.start(millisecondsWait);
    loop.exec();
}
void Window::clearButtonClicked()
{
    for (auto &x : m_cell)
    {
        if (x.collapsed)
            x.un();
    }
    return;
}
void Window::setValue(int s)
{
    m_delay = s;
    return;
}